@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2018 Fabian Groffen
+ * Copyright 2013-2021 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -33,10 +33,49 @@
 #ifdef HAVE_SSL
 #include <openssl/ssl.h>
 #include <openssl/err.h>
-#endif
-
 
 static char
+ssllisten(listener *lsnr)
+{
+	/* create a auto-negotiate context */
+	const SSL_METHOD *m = SSLv23_server_method();
+	lsnr->ctx = SSL_CTX_new(m);
+	if (lsnr->ctx == NULL) {
+		char *err = ERR_error_string(ERR_get_error(), NULL);
+		logerr("failed to create SSL context for listener on port %d: %s\n",
+				lsnr->port, err);
+		return 1;
+	}
+
+	/* load certificates */
+	if (SSL_CTX_use_certificate_file(
+				lsnr->ctx, lsnr->pemcert, SSL_FILETYPE_PEM) <= 0)
+	{
+		char *err = ERR_error_string(ERR_get_error(), NULL);
+		SSL_CTX_free(lsnr->ctx);
+		logerr("cannot load certificate from %s: %s\n", lsnr->pemcert, err);
+		return 1;
+	}
+	if (SSL_CTX_use_PrivateKey_file(
+				lsnr->ctx, lsnr->pemcert, SSL_FILETYPE_PEM) <= 0)
+	{
+		char *err = ERR_error_string(ERR_get_error(), NULL);
+		SSL_CTX_free(lsnr->ctx);
+		logerr("cannot load private key from %s: %s\n", lsnr->pemcert, err);
+		return 1;
+	}
+	if (!SSL_CTX_check_private_key(lsnr->ctx)) {
+		char *err = ERR_error_string(ERR_get_error(), NULL);
+		SSL_CTX_free(lsnr->ctx);
+		logerr("private key does not match public certificate in %s: %s\n",
+				lsnr->pemcert, err);
+		return 1;
+	}
+	return 0;
+}
+#endif
+
+static int
 bindlistenip(listener *lsnr, unsigned int backlog)
 {
 	int sock;
@@ -48,41 +87,8 @@ bindlistenip(listener *lsnr, unsigned int backlog)
 	int sockcur = 0;
 
 #ifdef HAVE_SSL
-	if ((lsnr->transport & ~0xFFFF) == W_SSL) {
-		/* create a auto-negotiate context */
-		const SSL_METHOD *m = SSLv23_server_method();
-		lsnr->ctx = SSL_CTX_new(m);
-		if (lsnr->ctx == NULL) {
-			char *err = ERR_error_string(ERR_get_error(), NULL);
-			logerr("failed to create SSL context for listener on port %d: %s\n",
-					lsnr->port, err);
-			return 1;
-		}
-
-		/* load certificates */
-		if (SSL_CTX_use_certificate_file(
-					lsnr->ctx, lsnr->pemcert, SSL_FILETYPE_PEM) <= 0)
-		{
-			char *err = ERR_error_string(ERR_get_error(), NULL);
-			SSL_CTX_free(lsnr->ctx);
-			logerr("cannot load certificate from %s: %s\n", lsnr->pemcert, err);
-			return 1;
-		}
-		if (SSL_CTX_use_PrivateKey_file(
-					lsnr->ctx, lsnr->pemcert, SSL_FILETYPE_PEM) <= 0)
-		{
-			char *err = ERR_error_string(ERR_get_error(), NULL);
-			SSL_CTX_free(lsnr->ctx);
-			logerr("cannot load private key from %s: %s\n", lsnr->pemcert, err);
-			return 1;
-		}
-		if (!SSL_CTX_check_private_key(lsnr->ctx)) {
-			char *err = ERR_error_string(ERR_get_error(), NULL);
-			SSL_CTX_free(lsnr->ctx);
-			logerr("private key does not match public certificate in %s: %s\n",
-					lsnr->pemcert, err);
-			return 1;
-		}
+	if ((lsnr->transport & ~0xFFFF) == W_SSL && ssllisten(lsnr)) {
+		return 1;
 	}
 #endif
 
@@ -165,12 +171,18 @@ bindlistenip(listener *lsnr, unsigned int backlog)
 	return 0;
 }
 
-static char
+static int
 bindlistenunix(listener *lsnr, unsigned int backlog)
 {
 	struct sockaddr_un srvr;
 	int sock;
 
+#ifdef HAVE_SSL
+	if ((lsnr->transport & ~0xFFFF) == W_SSL && ssllisten(lsnr)) {
+		return 1;
+	}
+#endif
+
 #ifndef PF_LOCAL
 # define PF_LOCAL PF_UNIX
 #endif
@@ -210,7 +222,7 @@ bindlistenunix(listener *lsnr, unsigned int backlog)
  * Open up sockets associated with listener.  Returns 0 when opening up
  * the listener succeeded, 1 otherwise.
  */
-char
+int
 bindlisten(listener *lsnr, unsigned int backlog)
 {
 	switch (lsnr->ctype) {
@@ -231,6 +243,10 @@ close_socks(listener *lsnr)
 	int i;
 	for (i = 0; lsnr->socks[i] != -1; i++)
 		close(lsnr->socks[i]);
+#ifdef HAVE_SSL
+	if ((lsnr->transport & ~0xFFFF) == W_SSL)
+		SSL_CTX_free(lsnr->ctx);
+#endif
 	logout("closed listener for %s %s:%u\n",
 			lsnr->ctype == CON_UDP ? "udp" : "tcp",
 			lsnr->ip == NULL ? "" : lsnr->ip, lsnr->port);
@@ -241,6 +257,10 @@ destroy_usock(listener *lsnr)
 {
 	close(lsnr->socks[0]);
 	unlink(lsnr->ip);
+#ifdef HAVE_SSL
+	if ((lsnr->transport & ~0xFFFF) == W_SSL)
+		SSL_CTX_free(lsnr->ctx);
+#endif
 	logout("removed listener for %s\n", lsnr->ip);
 }
 
